package hu.bme.cs.benchmark.mutex.db;

import hu.bme.cs.db.Edge;
import hu.bme.cs.db.Graph;
import hu.bme.cs.db.ID;
import hu.bme.cs.db.Node;
import hu.bme.cs.db.Pattern;
import hu.bme.cs.db.Rule;
import hu.bme.cs.db.TopPattern;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class Mutex {
    // TODO multiplicity based optimization theoretically
    // works for DB-based GT tools
    /**
     * Generates the database representation 
     * (i.e., the database schema) of the metamodel
     * @param conn
     * @return
     * @throws SQLException
     */
    public static Graph generateMetaModel(Connection conn) 
    	throws SQLException {
        Graph meta = new Graph();
        Node proc = new Node("process");
        Node res = new Node("resource");
        proc.setLabel("process");
        res.setLabel("resource");
        meta.addVertex(proc);
        meta.addVertex(res);
        Edge req = meta.addEdge(proc,res);
        Edge held_by = meta.addEdge(res,proc);
        Edge rel = meta.addEdge(res,proc);
        Edge next = meta.addEdge(proc,proc);
        Edge token = meta.addEdge(res,proc);
        Edge blocked = meta.addEdge(res,proc);
        req.setLabel("request");
        held_by.setLabel("held_by");
        rel.setLabel("release");
        next.setLabel("next");
        token.setLabel("token");
        blocked.setLabel("blocked");
        return meta;
    }

    /**
     * Generates the database representation 
     * (i.e., the views) of the GT rules
     * @param conn
     * @return
     * @throws SQLException
     */
    public static Map generateRules(Connection conn) 
    	throws SQLException {
        HashMap result = new HashMap();
        Node p,p1,p2,r,r1,r2;
        Edge rq,rq1,rq2,rl,n,n1,n2,n3,t,t1,t2,h,h1,h2,b,b1,b2;
        TopPattern lhsPattern = null;
        Pattern[] nacs = new Pattern[2];
        Graph lhs,rhs,nac;

        // Rule1 (new)
        lhs = new Graph();
        rhs = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        p = new Node("process");
        p1.setLabel("p1");
        p2.setLabel("p2");
        p.setLabel("p");
        lhs.addVertex(p1);
        lhs.addVertex(p2);
        rhs.addVertex(p1);
        rhs.addVertex(p2);
        rhs.addVertex(p);
        
        n1 = new Edge("next",p1,p2);
        n2 = new Edge("next",p1,p);
        n3 = new Edge("next",p,p2);
        n1.setLabel("n1");
        n2.setLabel("n2");
        n3.setLabel("n3");
        lhs.addEdge(n1);
        rhs.addEdge(n2);
        rhs.addEdge(n3);
        
        result.put("newRule",new Rule(conn,"newRule",lhs,rhs));
        
        // Rule2 (kill)
        lhs = new Graph();
        rhs = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        p = new Node("process");
        p1.setLabel("p1");
        p2.setLabel("p2");
        p.setLabel("p");
        lhs.addVertex(p1);
        lhs.addVertex(p2);
        lhs.addVertex(p);
        rhs.addVertex(p1);
        rhs.addVertex(p2);
        
        n1 = new Edge("next",p1,p2);
        n2 = new Edge("next",p1,p);
        n3 = new Edge("next",p,p2);
        n1.setLabel("n1");
        n2.setLabel("n2");
        n3.setLabel("n3");
        lhs.addEdge(n2);
        lhs.addEdge(n3);
        rhs.addEdge(n1);
        
        result.put("killRule",new Rule(conn,"killRule",lhs,rhs));
        
        // Rule3 (mount)
        lhs = new Graph();
        rhs = new Graph();
        
        p = new Node("process");
        r = new Node("resource");
        p.setLabel("p");
        r.setLabel("r");
        lhs.addVertex(p);
        rhs.addVertex(p);
        rhs.addVertex(r);
        
        t = new Edge("token",r,p);
        t.setLabel("t");
        rhs.addEdge(t);
        
        result.put("mountRule",new Rule(conn,"mountRule",lhs,rhs));
        
        // Rule4 (unmount)
        lhs = new Graph();
        rhs = new Graph();
        
        p = new Node("process");
        r = new Node("resource");
        p.setLabel("p");
        r.setLabel("r");
        lhs.addVertex(p);
        lhs.addVertex(r);
        rhs.addVertex(p);
        
        t = new Edge("token",r,p);
        t.setLabel("t");
        lhs.addEdge(t);
        
        result.put("unmountRule",new Rule(conn,"unmountRule",lhs,rhs));
        
        // Rule5 (pass)
        lhs = new Graph();
        rhs = new Graph();
        nac = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        r = new Node("resource");
        p1.setLabel("p1");
        p2.setLabel("p2");
        r.setLabel("r");
        lhs.addVertex(p1);
        lhs.addVertex(p2);
        lhs.addVertex(r);
        rhs.addVertex(p1);
        rhs.addVertex(p2);
        rhs.addVertex(r);
        nac.addVertex(p1);
        nac.addVertex(r);
        
        t1 = new Edge("token",r,p1);
        t2 = new Edge("token",r,p2);
        n = new Edge("next",p1,p2);
        rq = new Edge("request",p1,r);
        t1.setLabel("t1");
        t2.setLabel("t2");
        n.setLabel("n");
        rq.setLabel("rq");
        lhs.addEdge(n);
        lhs.addEdge(t1);
        rhs.addEdge(n);
        rhs.addEdge(t2);
        nac.addEdge(rq);
        
        nacs = new Pattern[1];
        nacs[0] = new Pattern(conn,"passRule_NAC",nac,null);
        lhsPattern = new TopPattern(conn,"passRule",lhs,nacs);
        result.put("passRule",new Rule(conn,"passRule",lhsPattern,rhs));
        
        // Rule6 (req)
        lhs = new Graph();
        rhs = new Graph();
        
        p = new Node("process");
        r1 = new Node("resource");
        r2 = new Node("resource");
        p.setLabel("p");
        r1.setLabel("r1");
        r2.setLabel("r2");
        lhs.addVertex(p);
        lhs.addVertex(r1);
        rhs.addVertex(p);
        rhs.addVertex(r1);
        
        rq1 = new Edge("request",p,r1);
        rq2 = new Edge("request",p,r2);
        h = new Edge("held_by",r1,p);
        rq1.setLabel("rq1");
        rq2.setLabel("rq2");
        h.setLabel("h");
        rhs.addEdge(rq1);
        
        nacs = new Pattern[2];
        nac = new Graph();
        nac.addVertex(p);
        nac.addVertex(r2);
        nac.addEdge(rq2);
        nacs[0] = new Pattern(conn,"requestRule_NAC1",nac,null);
        
        nac = new Graph();
        nac.addVertex(p);
        nac.addVertex(r1);
        nac.addEdge(h);
        nacs[1] = new Pattern(conn,"requestRule_NAC2",nac,null);

        lhsPattern = new TopPattern(conn,"requestRule",lhs,nacs);
        result.put("requestRule",new Rule(conn,"requestRule",lhsPattern,rhs));

        // Rule7 (take)
        lhs = new Graph();
        rhs = new Graph();
        
        p = new Node("process");
        r = new Node("resource");
        p.setLabel("p");
        r.setLabel("r");
        lhs.addVertex(p);
        lhs.addVertex(r);
        rhs.addVertex(p);
        rhs.addVertex(r);
        
        t = new Edge("token",r,p);
        rq = new Edge("request",p,r);
        h = new Edge("held_by",r,p);
        t.setLabel("t");
        h.setLabel("h");
        rq.setLabel("rq");
        lhs.addEdge(t);
        lhs.addEdge(rq);
        rhs.addEdge(h);
        
        result.put("takeRule",new Rule(conn,"takeRule",lhs,rhs));
        
        // Rule8 (rel)
        lhs = new Graph();
        rhs = new Graph();
        nac = new Graph();
        
        p = new Node("process");
        r1 = new Node("resource");
        r2 = new Node("resource");
        p.setLabel("p");
        r1.setLabel("r1");
        r2.setLabel("r2");
        lhs.addVertex(p);
        lhs.addVertex(r1);
        rhs.addVertex(p);
        rhs.addVertex(r1);
        nac.addVertex(p);
        nac.addVertex(r2);
        
        rq = new Edge("request",p,r2);
        h = new Edge("held_by",r1,p);
        rl = new Edge("release",r1,p);
        rl.setLabel("rl");
        h.setLabel("h");
        rq.setLabel("rq");
        lhs.addEdge(h);
        rhs.addEdge(rl);
        nac.addEdge(rq);
        
        nacs = new Pattern[1];
        nacs[0] = new Pattern(conn,"releaseRule_NAC",nac,null);
        lhsPattern = new TopPattern(conn,"releaseRule",lhs,nacs);
        result.put("releaseRule",new Rule(conn,"releaseRule",lhsPattern,rhs));
        
        // Rule9 (give)
        lhs = new Graph();
        rhs = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        r = new Node("resource");
        p1.setLabel("p1");
        p2.setLabel("p2");
        r.setLabel("r");
        lhs.addVertex(p1);
        lhs.addVertex(p2);
        lhs.addVertex(r);
        rhs.addVertex(p1);
        rhs.addVertex(p2);
        rhs.addVertex(r);
        
        t = new Edge("token",r,p2);
        n = new Edge("next",p1,p2);
        rl = new Edge("release",r,p1);
        t.setLabel("t");
        n.setLabel("n");
        rl.setLabel("rl");
        lhs.addEdge(n);
        lhs.addEdge(rl);
        rhs.addEdge(n);
        rhs.addEdge(t);
        
        result.put("giveRule",new Rule(conn,"giveRule",lhs,rhs));        

        // Rule10 (blocked)
        lhs = new Graph();
        rhs = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        r = new Node("resource");
        p1.setLabel("p1");
        p2.setLabel("p2");
        r.setLabel("r");
        lhs.addVertex(p1);
        lhs.addVertex(p2);
        lhs.addVertex(r);
        rhs.addVertex(p1);
        rhs.addVertex(p2);
        rhs.addVertex(r);
        
        b = new Edge("blocked",r,p1);
        rq = new Edge("request",p1,r);
        h = new Edge("held_by",r,p2);
        b.setLabel("b");
        rq.setLabel("rq");
        h.setLabel("h");
        lhs.addEdge(rq);
        lhs.addEdge(h);
        rhs.addEdge(rq);
        rhs.addEdge(h);
        rhs.addEdge(b);
        
        result.put("blockedRule",new Rule(conn,"blockedRule",lhs,rhs));
        
        // Rule11 (waiting)
        lhs = new Graph();
        rhs = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        r1 = new Node("resource");
        r2 = new Node("resource");
        p1.setLabel("p1");
        p2.setLabel("p2");
        r1.setLabel("r1");
        r2.setLabel("r2");
        lhs.addVertex(p1);
        lhs.addVertex(p2);
        lhs.addVertex(r1);
        lhs.addVertex(r2);
        rhs.addVertex(p1);
        rhs.addVertex(p2);
        rhs.addVertex(r1);
        rhs.addVertex(r2);
        
        b1 = new Edge("blocked",r2,p1);
        b2 = new Edge("blocked",r2,p2);
        rq = new Edge("request",p2,r1);
        h = new Edge("held_by",r1,p1);
        b1.setLabel("b1");
        b2.setLabel("b2");
        rq.setLabel("rq");
        h.setLabel("h");
        lhs.addEdge(rq);
        lhs.addEdge(h);
        lhs.addEdge(b1);
        rhs.addEdge(rq);
        rhs.addEdge(h);
        rhs.addEdge(b2);
        
        result.put("waitingRule",new Rule(conn,"waitingRule",lhs,rhs));
        
        // Rule12 (ignore)
        lhs = new Graph();
        rhs = new Graph();
        nac = new Graph();
        
        p = new Node("process");
        r1 = new Node("resource");
        r2 = new Node("resource");
        p.setLabel("p");
        r1.setLabel("r1");
        r2.setLabel("r2");
        lhs.addVertex(p);
        lhs.addVertex(r1);
        rhs.addVertex(p);
        rhs.addVertex(r1);
        nac.addVertex(p);
        nac.addVertex(r2);
        
        b = new Edge("blocked",r1,p);
        h = new Edge("held_by",r2,p);
        b.setLabel("b");
        h.setLabel("h");
        lhs.addEdge(b);
        nac.addEdge(h);
        
        nacs = new Pattern[1];
        nacs[0] = new Pattern(conn,"ignoreRule_NAC",nac,null);
        lhsPattern = new TopPattern(conn,"ignoreRule",lhs,nacs);
        result.put("ignoreRule",new Rule(conn,"ignoreRule",lhsPattern,rhs));
        
        // Rule13 (unlock)
        lhs = new Graph();
        rhs = new Graph();
        nac = new Graph();
        
        p = new Node("process");
        r = new Node("resource");
        p.setLabel("p");
        r.setLabel("r");
        lhs.addVertex(p);
        lhs.addVertex(r);
        rhs.addVertex(p);
        rhs.addVertex(r);
        
        b = new Edge("blocked",r,p);
        h = new Edge("held_by",r,p);
        rl = new Edge("release",r,p);
        b.setLabel("b");
        h.setLabel("h");
        rl.setLabel("rl");
        lhs.addEdge(b);
        lhs.addEdge(h);
        rhs.addEdge(rl);
        
        result.put("unlockRule",new Rule(conn,"unlockRule",lhs,rhs));

        // Rule 6/a (requestStar)
        lhs = new Graph();
        rhs = new Graph();
        nac = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        r1 = new Node("resource");
        r2 = new Node("resource");
        p1.setLabel("p1");
        p2.setLabel("p2");
        r1.setLabel("r1");
        r2.setLabel("r2");
        lhs.addVertex(p1);
        lhs.addVertex(r1);
        lhs.addVertex(p2);
        lhs.addVertex(r2);
        rhs.addVertex(p1);
        rhs.addVertex(r1);
        rhs.addVertex(p2);
        rhs.addVertex(r2);
        nac.addVertex(p1);
        nac.addVertex(r2);
        
        n = new Edge("next",p2,p1);
        h1 = new Edge("held_by",r1,p1);
        h2 = new Edge("held_by",r2,p2);
        rq1 = new Edge("request",p1,r2);
        rq2 = new Edge("request",p1,r2);
        n.setLabel("n");
        h1.setLabel("h1");
        h2.setLabel("h2");
        rq1.setLabel("rq1");
        rq2.setLabel("rq2");
        lhs.addEdge(n);
        lhs.addEdge(h1);
        lhs.addEdge(h2);
        rhs.addEdge(n);
        rhs.addEdge(h1);
        rhs.addEdge(h2);
        rhs.addEdge(rq1);
        nac.addEdge(rq2);
        
        nacs = new Pattern[1];
        nacs[0] = new Pattern(conn,"requestStarRule_NAC",nac,null);

        lhsPattern = new TopPattern(conn,"requestStarRule",lhs,nacs);
        result.put("requestStarRule",
                new Rule(conn,"requestStarRule",lhsPattern,rhs));

        // Rule 8/a (releaseStar)
        lhs = new Graph();
        rhs = new Graph();
        
        p1 = new Node("process");
        p2 = new Node("process");
        r1 = new Node("resource");
        r2 = new Node("resource");
        p1.setLabel("p1");
        p2.setLabel("p2");
        r1.setLabel("r1");
        r2.setLabel("r2");
        lhs.addVertex(p1);
        lhs.addVertex(r1);
        lhs.addVertex(p2);
        lhs.addVertex(r2);
        rhs.addVertex(p1);
        rhs.addVertex(r1);
        rhs.addVertex(p2);
        rhs.addVertex(r2);
        nac.addVertex(p1);
        
        h1 = new Edge("held_by",r1,p2);
        h2 = new Edge("held_by",r2,p2);
        rq = new Edge("request",p1,r1);
        rl = new Edge("release",r1,p2);
        h1.setLabel("h1");
        h2.setLabel("h2");
        rq.setLabel("rq");
        rl.setLabel("rl");

        lhs.addEdge(h1);
        lhs.addEdge(h2);
        lhs.addEdge(rq);
        rhs.addEdge(rq);
        rhs.addEdge(rl);
        rhs.addEdge(h2);

        result.put("releaseStarRule",
                new Rule(conn,"releaseStarRule",lhs,rhs));
        
        // Rule 6/b (requestSimple)
        lhs = new Graph();
        rhs = new Graph();
        nac = new Graph();
        
        p = new Node("process");
        r = new Node("resource");
        p.setLabel("p");
        r.setLabel("r");
        lhs.addVertex(p);
        lhs.addVertex(r);
        rhs.addVertex(p);
        rhs.addVertex(r);
        nac.addVertex(p);
        nac.addVertex(r);
        
        t = new Edge("token",r,p);
        rq1 = new Edge("request",p,r);
        rq2 = new Edge("request",p,r);
        t.setLabel("t");
        rq1.setLabel("rq1");
        rq2.setLabel("rq2");
        lhs.addEdge(t);
        rhs.addEdge(t);
        rhs.addEdge(rq1);
        nac.addEdge(rq2);
        
        nacs = new Pattern[1];
        nacs[0] = new Pattern(conn,"requestSimpleRule_NAC",nac,null);

        lhsPattern = new TopPattern(conn,"requestSimpleRule",lhs,nacs);
        result.put("requestSimpleRule",
                new Rule(conn,"requestSimpleRule",lhsPattern,rhs));
        return result;
    }
    
    public static void generateModel(Connection conn, Graph g)
    	throws SQLException {
        // TODO Top-down traversal of the type hierarchy
        PreparedStatement insert = conn.prepareStatement(
                "INSERT INTO object VALUES (?)");
        Statement stmt = conn.createStatement();
        HashMap map = new HashMap();
        
        Iterator i = g.vertexSet().iterator();
        while(i.hasNext()) {
            Node elem = (Node) i.next();
            int id = ID.getNext();
            map.put(elem,new Integer(id));
            insert.setInt(1,((Integer) map.get(elem)).intValue());
            insert.executeUpdate();
            stmt.executeUpdate("INSERT INTO " + elem.getType() + " VALUES ('" 
                    + ((Integer) map.get(elem)).intValue() + "')");
        }
        i = g.edgeSet().iterator();
        while(i.hasNext()) {
            Edge elem = (Edge) i.next();
            int id = ID.getNext();
            map.put(elem,new Integer(id));
            insert.setInt(1,((Integer) map.get(elem)).intValue());
            insert.executeUpdate();
            stmt.executeUpdate("INSERT INTO " + elem.getType() + " VALUES ('" 
                    + ((Integer) map.get(elem)).intValue() + "','" 
                    + ((Integer) map.get((Node) elem.getSource())).intValue() + "','" 
                    + ((Integer) map.get((Node) elem.getTarget())).intValue() + "')");
        }
        conn.commit();
    }
}
