/*
 * bv2 project
 * Created on Feb 10, 2005
 *
 * Copyright (C) 2005
 * Gergely Varro
 * Department of Computer Science and Information Theory
 * Budapest University of Technology and Economics
 * Budapest, Hungary
 */
package hu.bme.cs.benchmark.mutex.db;

import hu.bme.mit.viatra.core.*;
import hu.bme.mit.viatra.core.simple.SimpleModelSpace;
import hu.bme.mit.viatra.errors.VPMCoreException;
import hu.bme.mit.viatra.errors.VPMCoreNullParameterException;
import hu.bme.mit.viatra.errors.VPMRuntimeException;
import hu.bme.mit.logger.SimpleLogger;
/**
 * @author gergo
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class VPMMutex {
    public static void main(String[] args) 
    	throws VPMCoreNullParameterException,VPMCoreException,VPMRuntimeException {
        final int N = 5;
        IModelSpace ms = new SimpleModelSpace();
        // inicializalas eleg furcsa (null-os elszallassal all meg)
        // getRoot()-nal nincs Exception handling
        // IEntity -> Entity
        // Logger-nel jo lenne valami Apache-ot hasznalni
        ms.init(new SimpleLogger(),null);
        
        IModelManager m = ms.getModelManager();
                
        // Metamodel
        IEntity metamodel = m.newEntity("metamodel");
        IEntity process = m.newEntity("Process",metamodel);
        IEntity resource = m.newEntity("Resource",metamodel);
        IRelation next = m.newRelation("next",process,process,metamodel);
        IRelation token = m.newRelation("token",resource,process,metamodel);
        IRelation request = m.newRelation("request",process,resource,metamodel);
        IRelation held_by = m.newRelation("held_by",resource,process,metamodel);
        IRelation release = m.newRelation("release",resource,process,metamodel);
        IRelation blocked = m.newRelation("blocked",resource,process,metamodel);
        
        // Rules
        IEntity lhs,rhs,nac;
        IEntity p,p1,p2,r,r1,r2;
        IRelation rq,rq1,rq2,rl,n,n1,n2,n3,t,t1,t2,h,h1,h2,b,b1,b2;
        
        // newRule
        IEntity newRule = m.newEntity("newRule");
        lhs = m.newEntity(newRule);
        rhs = m.newEntity(newRule);
        
        p1 = m.newEntity("p1",lhs);
        m.newContainment(rhs,p1);
        m.newInstanceOf(process,p1);
        p2 = m.newEntity("p2",lhs);
        m.newContainment(rhs,p2);
        m.newInstanceOf(process,p2);
        
        n = m.newRelation("n",p1,p2,lhs);
        m.newInstanceOf(next,n);
        
        p = m.newEntity("p",rhs);
        m.newInstanceOf(process,p);
        n1 = m.newRelation("n1",p1,p,rhs);
        m.newInstanceOf(next,n1);
        n2 = m.newRelation("n2",p,p2,rhs);
        m.newInstanceOf(next,n2);
        
        // Rule2 (kill)
        IEntity killRule = m.newEntity("killRule");
        lhs = m.newEntity(killRule);
        rhs = m.newEntity(killRule);
        
        p1 = m.newEntity("p1",lhs);
        m.newContainment(rhs,p1);
        m.newInstanceOf(process,p1);
        p2 = m.newEntity("p2",lhs);
        m.newContainment(rhs,p2);
        m.newInstanceOf(process,p2);
        
        p = m.newEntity("p",lhs);
        m.newInstanceOf(process,p);
        n2 = m.newRelation("n2",p1,p,lhs);
        m.newInstanceOf(next,n2);
        n3 = m.newRelation("n3",p,p2,lhs);
        m.newInstanceOf(next,n3);
        
        n1 = m.newRelation("n1",p1,p2,rhs);
        
        // Rule3 (mount)
        IEntity mountRule = m.newEntity("mountRule");
        lhs = m.newEntity(mountRule);
        rhs = m.newEntity(mountRule);
        
        p = m.newEntity("p",lhs);
        m.newContainment(rhs,p);
        m.newInstanceOf(process,p);
        
        r = m.newEntity("r",rhs);
        m.newInstanceOf(resource,r);
        t = m.newRelation("t",r,p,rhs);
        m.newInstanceOf(token,t);
        
        // Rule4 (unmount)
        IEntity unmountRule = m.newEntity("unmountRule");
        lhs = m.newEntity(unmountRule);
        rhs = m.newEntity(unmountRule);
        
        p = m.newEntity("p",lhs);
        m.newContainment(rhs,p);
        m.newInstanceOf(process,p);
        
        r = m.newEntity("r",lhs);
        m.newInstanceOf(resource,r);
        t = m.newRelation("t",r,p,lhs);
        
        // Rule5 (pass)
        IEntity passRule = m.newEntity("passRule");
        lhs = m.newEntity(passRule);
        rhs = m.newEntity(passRule);
        nac = m.newEntity(passRule);
        
        p1 = m.newEntity("p1",lhs);
        m.newContainment(rhs,p1);
        m.newContainment(nac,p1);
        m.newInstanceOf(process,p1);
        p2 = m.newEntity("p2",lhs);
        m.newContainment(rhs,p2);
        m.newInstanceOf(process,p2);
        r = m.newEntity("r",lhs);
        m.newContainment(rhs,r);
        m.newContainment(nac,r);
        m.newInstanceOf(resource,r);
        n = m.newRelation("n",p1,p2,lhs);
        m.newContainment(rhs,n);
        m.newInstanceOf(next,n);
        
        t1 = m.newRelation("t1",r,p1,lhs);
        m.newInstanceOf(token,t1);
        
        t2 = m.newRelation("t2",r,p2,rhs);
        m.newInstanceOf(token,t2);
        
        rq = m.newRelation("rq",p1,r,nac);
        m.newInstanceOf(request,rq);
        
        // Rule6 (req)
        IEntity requestRule = m.newEntity("requestRule");
        lhs = m.newEntity(requestRule);
        rhs = m.newEntity(requestRule);
        nac = m.newEntity(requestRule);
        IEntity nacExtra = m.newEntity(requestRule);
        
        p = m.newEntity("p",lhs);
        m.newContainment(rhs,p);
        m.newContainment(nac,p);
        m.newContainment(nacExtra,p);
        m.newInstanceOf(process,p);
        r1 = m.newEntity("r1",lhs);
        m.newContainment(rhs,r1);
        m.newContainment(nacExtra,r1);
        m.newInstanceOf(resource,r1);
        
        rq1 = m.newRelation("rq1",p,r1,rhs);
        m.newInstanceOf(request,rq1);
        
        r2 = m.newEntity("r2",nac);
        m.newInstanceOf(resource,r2);
        rq2 = m.newRelation("rq2",p,r2,nac);
        m.newInstanceOf(request,rq2);
        
        h = m.newRelation("h",r1,p,nacExtra);
        m.newInstanceOf(held_by,h);
        
        // Rule7 (take)
        IEntity takeRule = m.newEntity("takeRule");
        lhs = m.newEntity(takeRule);
        rhs = m.newEntity(takeRule);
        
        p = m.newEntity("p",lhs);
        m.newContainment(rhs,p);
        m.newInstanceOf(process,p);
        r = m.newEntity("r",lhs);
        m.newContainment(rhs,r);
        m.newInstanceOf(resource,r);
        
        t = m.newRelation("t",r,p,lhs);
        m.newInstanceOf(token,t);
        rq = m.newRelation("rq",p,r,lhs);
        m.newInstanceOf(request,rq);
        
        h = m.newRelation("h",r,p,rhs);
        m.newInstanceOf(held_by,h);
        
        // Rule8 (rel)
        IEntity releaseRule = m.newEntity("releaseRule");
        lhs = m.newEntity(releaseRule);
        rhs = m.newEntity(releaseRule);
        nac = m.newEntity(releaseRule);
        
        p = m.newEntity("p",lhs);
        m.newContainment(rhs,p);
        m.newContainment(nac,p);
        m.newInstanceOf(process,p);
        r1 = m.newEntity("r1",lhs);
        m.newContainment(rhs,r1);
        m.newInstanceOf(resource,r1);
        
        h = m.newRelation("h",r1,p,lhs);
        m.newInstanceOf(held_by,h);
        
        r2 = m.newEntity("r2",rhs);
        m.newInstanceOf(resource,r2);
        rl = m.newRelation("rl",r1,p,rhs);
        m.newInstanceOf(release,rl);
        
        rq = m.newRelation("rq",p,r2,nac);
        m.newInstanceOf(request,rq);
        
        // Rule9 (give)
        IEntity giveRule = m.newEntity("giveRule");
        lhs = m.newEntity(giveRule);
        rhs = m.newEntity(giveRule);
        
        p1 = m.newEntity("p1",lhs);
        m.newContainment(rhs,p1);
        m.newInstanceOf(process,p1);
        p2 = m.newEntity("p2",lhs);
        m.newContainment(rhs,p2);
        m.newInstanceOf(process,p2);
        r = m.newEntity("r",lhs);
        m.newContainment(rhs,r);
        m.newInstanceOf(resource,r);
        n = m.newRelation("n",p1,p2,lhs);
        m.newContainment(rhs,n);
        m.newInstanceOf(next,n);
        
        rl = m.newRelation("rl",r,p1,lhs);
        m.newInstanceOf(release,rl);
        
        t = m.newRelation("t",r,p2,rhs);
        m.newInstanceOf(token,t);
        
        // Rule10 (blocked)
        IEntity blockedRule = m.newEntity("blockedRule");
        lhs = m.newEntity(blockedRule);
        rhs = m.newEntity(blockedRule);
        
        p1 = m.newEntity("p1",lhs);
        m.newContainment(rhs,p1);
        m.newInstanceOf(process,p1);
        p2 = m.newEntity("p2",lhs);
        m.newContainment(rhs,p2);
        m.newInstanceOf(process,p2);
        r = m.newEntity("r",lhs);
        m.newContainment(rhs,r);
        m.newInstanceOf(resource,r);
        rq = m.newRelation("rq",p2,r,lhs);
        m.newContainment(rhs,rq);
        m.newInstanceOf(request,rq);
        h = m.newRelation("h",r,p1,lhs);
        m.newContainment(rhs,h);
        m.newInstanceOf(held_by,h);
        
        b = m.newRelation("b",r,p2,rhs);
        m.newInstanceOf(blocked,b);
        
        // Rule11 (waiting)
        IEntity waitingRule = m.newEntity("waitingRule");
        lhs = m.newEntity(waitingRule);
        rhs = m.newEntity(waitingRule);
        
        p1 = m.newEntity("p1",lhs);
        m.newContainment(rhs,p1);
        m.newInstanceOf(process,p1);
        p2 = m.newEntity("p2",lhs);
        m.newContainment(rhs,p2);
        m.newInstanceOf(process,p2);
        r1 = m.newEntity("r1",lhs);
        m.newContainment(rhs,r1);
        m.newInstanceOf(resource,r1);
        r2 = m.newEntity("r2",lhs);
        m.newContainment(rhs,r2);
        m.newInstanceOf(resource,r2);
        rq = m.newRelation("rq",p2,r1,lhs);
        m.newContainment(rhs,rq);
        m.newInstanceOf(request,rq);
        h = m.newRelation("h",r1,p1,lhs);
        m.newContainment(rhs,h);
        m.newInstanceOf(held_by,h);
        
        b1 = m.newRelation("b1",r2,p1,lhs);
        m.newInstanceOf(blocked,b1);
        
        b2 = m.newRelation("b2",r2,p2,rhs);
        m.newInstanceOf(blocked,b2);
        
        // Rule12 (ignore)
        IEntity ignoreRule = m.newEntity("ignoreRule");
        lhs = m.newEntity(ignoreRule);
        rhs = m.newEntity(ignoreRule);
        nac = m.newEntity(ignoreRule);
        
        p = m.newEntity("p",lhs);
        m.newContainment(rhs,p);
        m.newContainment(nac,p);
        m.newInstanceOf(process,p);
        r1 = m.newEntity("r1",lhs);
        m.newContainment(rhs,r1);
        m.newInstanceOf(resource,r1);
        
        b = m.newRelation("b",r1,p,lhs);
        m.newInstanceOf(blocked,b);
        
        r2 = m.newEntity("r2",nac);
        m.newInstanceOf(resource,r2);
        h = m.newRelation("h",r2,p,nac);
        m.newInstanceOf(held_by,h);
        
        // Rule13 (unlock)
        IEntity unlockRule = m.newEntity("unlockRule");
        lhs = m.newEntity(unlockRule);
        rhs = m.newEntity(unlockRule);
        
        p = m.newEntity("p",lhs);
        m.newContainment(rhs,p);
        m.newInstanceOf(process,p);
        r = m.newEntity("r",lhs);
        m.newContainment(rhs,r);
        m.newInstanceOf(resource,r);
        
        b = m.newRelation("b",r,p,lhs);
        m.newInstanceOf(blocked,b);
        h = m.newRelation("h",r,p,lhs);
        m.newInstanceOf(held_by,h);
        
        rl = m.newRelation("rl",r,p,rhs);
        m.newInstanceOf(release,rl);
        
        // Initial model STS
        IEntity modelSTS = m.newEntity("initSTS");
        p1 = m.newEntity("p1",modelSTS);
        m.newInstanceOf(process,p1);
        p2 = m.newEntity("p2",modelSTS);
        m.newInstanceOf(process,p2);
        n1 = m.newRelation("n1",p1,p2,modelSTS);
        m.newInstanceOf(next,n1);
        n2 = m.newRelation("n2",p2,p1,modelSTS);
        m.newInstanceOf(next,n2);
        
        // Initial model LTS
        IEntity modelLTS = m.newEntity("initLTS");
        IEntity p0 = m.newEntity(modelLTS);
        m.newInstanceOf(process,p0);
        IEntity r0 = m.newEntity(modelLTS);
        m.newInstanceOf(resource,r0);
        IRelation t0 = m.newRelation(r0,p0,modelLTS);
        
        IEntity first = p0;
        IEntity old = p0;
        for(int i = 0; i < N; i++) {
            p0 = m.newEntity(modelLTS);
            m.newInstanceOf(process,p0);
            r0 = m.newEntity(modelLTS);
            m.newInstanceOf(resource,r0);
            t0 = m.newRelation(r0,p0,modelLTS);
            n = m.newRelation(old,p0,modelLTS);
            m.newInstanceOf(next,n);
            old = p0;
        }
        m.newInstanceOf(next,m.newRelation(old,first,modelLTS));
     }
}
